(function() {
  var Freakset;
  Freakset = function() {
    function Freakset(opts) {
      opts || (opts = {});
      this.stack = opts.stack || (opts.stack = []);
      this.store = opts.store || (opts.store = {});
      this.outerFreakset = opts.outerFreakset || (opts.outerFreakset = void 0);
      this.resolvedCallback = opts.resolvedCallback || (opts.resolvedCallback = void 0);
      this.runInParallel = opts.runInParallel || (opts.runInParallel = void 0);
      this.running = false;
      this.innerFreaksets = [];
    }
    Freakset.build = function(opts) {
      return new Freakset(opts || (opts = {}));
    };
    Freakset.prototype.isRunning = function() {
      var innerFreakset, _i, _len, _ref;
      if (this.running === true) {
        return true;
      }
      if (this.innerFreaksets.length !== 0) {
        _ref = this.innerFreaksets;
        for (_i = 0, _len = _ref.length; _i < _len; _i++) {
          innerFreakset = _ref[_i];
          if (innerFreakset.isRunning()) {
            return true;
          }
        }
      }
      return false;
    };
    Freakset.prototype.root = function() {
      if (this.outerFreakset != null) {
        return this.outerFreakset.root();
      } else {
        return this;
      }
    };
    Freakset.prototype.step = function(namedScope, fn) {
      if (this.runInParallel != null) {
        this.parallelStepCount || (this.parallelStepCount = 0);
        this.parallelStepCount += 1;
      }
      return this.pushToStack(this.stack, {
        type: "step",
        scope: namedScope,
        fn: fn
      });
    };
    Freakset.prototype.group = function(namedScope, opts, fn) {
      var innerFreakset;
      if (typeof opts === "function") {
        fn = opts;
        opts = void 0;
      }
      opts || (opts = {});
      opts.parallel || (opts.parallel = false);
      opts.inFront || (opts.inFront = false);
      innerFreakset = Freakset.build({
        store: this.store,
        outerFreakset: this,
        runInParallel: opts.parallel
      });
      this.pushToStack(this.stack, {
        type: "group",
        scope: namedScope,
        fn: fn,
        innerFreakset: innerFreakset,
        parallel: opts.parallel
      }, opts);
      this.innerFreaksets.push(innerFreakset);
      fn(innerFreakset);
      return innerFreakset;
    };
    Freakset.prototype.rescue = function(namedScope, errFn, fn) {
      var innerFreakset;
      innerFreakset = Freakset.build({
        store: this.store,
        outerFreakset: this
      });
      this.pushToStack(this.stack, {
        type: "rescue",
        scope: namedScope,
        fn: fn,
        errFn: errFn,
        innerFreakset: innerFreakset
      });
      this.innerFreaksets.push(innerFreakset);
      fn(innerFreakset);
      return innerFreakset;
    };
    Freakset.prototype.guard = function(namedScope, fn) {
      return this.pushToStack(this.stack, {
        type: "guard",
        scope: namedScope,
        fn: fn
      });
    };
    Freakset.prototype.next = function() {
      var actor;
      this.running = true;
      actor = this.stack.shift();
      if (actor == null) {
        if (this.runInParallel != null) {
          if ((this.parallelStepCount != null) && this.parallelStepCount === 0) {
            return this.finish();
          } else {
            return;
          }
        } else {
          return this.finish();
        }
      }
      switch (actor.type) {
        case 'group':
          return this.runGroupActor(actor);
        case 'rescue':
          return this.runRescueActor(actor);
        case 'step':
          return this.runStepActor(actor);
        case 'guard':
          return this.runGuardActor(actor);
      }
    };
    Freakset.prototype.commit = function() {
      if ((this.runInParallel != null) && this.parallelStepCount) {
        this.parallelStepCount -= 1;
      }
      return this.next();
    };
    Freakset.prototype.runGroupActor = function(actor) {
      return this.buildAndRunFreakset({
        stack: actor.innerFreakset.stack,
        store: this.store,
        outerFreakset: this,
        runInParallel: actor.parallel
      });
    };
    Freakset.prototype.runRescueActor = function(actor) {
      try {
        return this.buildAndRunFreakset({
          stack: actor.innerFreakset.stack,
          store: this.store,
          outerFreakset: this
        });
      } catch (e) {
        actor.errFn(e, this, actor.innerFreakset);
        return this.commit();
      }
    };
    Freakset.prototype.runStepActor = function(actor) {
      actor.fn(this);
      if (this.runInParallel != null) {
        return this.next();
      }
    };
    Freakset.prototype.runGuardActor = function(actor) {
      var success;
      success = actor.fn(this);
      if (success === true) {
        return this.commit();
      } else {
        return this.finish();
      }
    };
    Freakset.prototype.compile = function(store, resolvedCallback) {
      if (resolvedCallback != null) {
        this.resolvedCallback = resolvedCallback;
      }
      store || (store = {});
      return this.buildAndRunFreakset({
        stack: this.stack,
        store: store,
        resolvedCallback: resolvedCallback
      });
    };
    Freakset.prototype.finish = function() {
      this.running = false;
      this.resolveRoot();
      if (this.outerFreakset) {
        this.outerFreakset.commit();
      }
      return true;
    };
    Freakset.prototype.resolveRoot = function() {
      var root;
      root = this.root();
      if (!root.isRunning()) {
        if (root.resolvedCallback != null) {
          return root.resolvedCallback.call(root);
        }
      }
    };
    Freakset.prototype.buildAndRunFreakset = function(opts) {
      var freakset;
      if (opts.stack != null) {
        opts.stack = opts.stack.slice();
      }
      freakset = Freakset.build(opts);
      freakset.next();
      return freakset;
    };
    Freakset.prototype.pushToStack = function(stack, object, opts) {
      opts || (opts = {});
      if (!opts.inFront && this.isRunning()) {
        opts.inFront = true;
      } else {
        opts.inFront || (opts.inFront = false);
      }
      if (opts.inFront === true) {
        return stack.unshift(object);
      } else {
        return stack.push(object);
      }
    };
    return Freakset;
  }();
  exports.Freakset = Freakset;
}).call(this);
