class Freakset

  @build: (opts) ->
    new Freakset(opts ||= {})

  constructor: (opts) ->
    opts ||= {}
    @stack            = opts.stack ||= []
    @store            = opts.store ||= {}
    @outerFreakset    = opts.outerFreakset ||= undefined
    @resolvedCallback = opts.resolvedCallback ||= undefined
    @runInParallel    = opts.runInParallel ||= undefined

    @running = no
    @innerFreaksets = []


  isRunning: ->
    return true if @running == yes

    if @innerFreaksets.length != 0
      for innerFreakset in @innerFreaksets
        if innerFreakset.isRunning()
          return true

    return false

  root: ->
    if @outerFreakset? then @outerFreakset.root() else @

  step: (namedScope, fn) ->
    if @runInParallel?
      @parallelStepCount ||= 0
      @parallelStepCount += 1

    @pushToStack @stack, { type: "step", scope: namedScope, fn: fn }


  group: (namedScope, opts, fn) ->
    if typeof opts == "function"
      fn = opts
      opts = undefined

    opts ||= {}
    opts.parallel ||= no
    opts.inFront ||= no

    innerFreakset = Freakset.build( store: @store, outerFreakset: @, runInParallel: opts.parallel)
    @pushToStack @stack, { type: "group", scope: namedScope, fn: fn, innerFreakset: innerFreakset, parallel: opts.parallel }, opts
    @innerFreaksets.push(innerFreakset)
    fn(innerFreakset)
    return innerFreakset


  rescue: (namedScope, errFn, fn) ->
    # Pass the store and this as outerFreakset to the innerFreakset
    innerFreakset = Freakset.build( store: @store, outerFreakset: @ )
    @pushToStack @stack, { type: "rescue", scope: namedScope, fn: fn, errFn: errFn, innerFreakset: innerFreakset }
    @innerFreaksets.push(innerFreakset)
    fn(innerFreakset)
    return innerFreakset


  guard: (namedScope, fn) ->
    @pushToStack @stack, { type: "guard", scope: namedScope, fn: fn }

  next: ->
    @running = yes

    actor = @stack.shift()

    unless actor?
      if @runInParallel?
        if @parallelStepCount? && @parallelStepCount == 0
          return @finish()
        else
          return
      else
        return @finish()

    switch actor.type
      when 'group'
        @runGroupActor(actor)

      when 'rescue'
        @runRescueActor(actor)

      when 'step'
        @runStepActor(actor)

      when 'guard'
        @runGuardActor(actor)

  commit: ->
    # Decrease the finished step count in parallel mode
    if @runInParallel? && @parallelStepCount
      @parallelStepCount -= 1

    @next()

  runGroupActor: (actor) ->
    @buildAndRunFreakset(stack: actor.innerFreakset.stack, store: @store, outerFreakset: @, runInParallel: actor.parallel)


  runRescueActor: (actor) ->
    try
      @buildAndRunFreakset(stack: actor.innerFreakset.stack, store: @store, outerFreakset: @)
    catch e
      # TODO: We might want to throw a Freakset exception if a errFn is not defined
      actor.errFn(e, @, actor.innerFreakset)
      @commit()


  runStepActor: (actor) ->
    # Run the actor in the context of this
    # Note to myself: We cannot automatically #commit after running the fn because
    # if the step contains running more closures it may not be finished at this point.
    actor.fn(@)

    # Experimental: If the step shall be run in parallel, just #commit() to start the next step..
    if @runInParallel?
      @next()


  runGuardActor: (actor) ->
    # Run the actor in the context of this
    success = actor.fn(@)
    if success is yes
      @commit()
    else
      @finish()


  compile: (store, resolvedCallback) ->
    # Attach resolvedCallback
    if resolvedCallback?
      @resolvedCallback = resolvedCallback

    # Prepare store if its undefined
    store ||= {}

    @buildAndRunFreakset(stack: @stack, store: store, resolvedCallback: resolvedCallback)

  finish: () ->
    @running = no
    @resolveRoot()
    @outerFreakset.commit() if @outerFreakset
    return true

  resolveRoot: ->
    root = @root()
    if !root.isRunning()
      root.resolvedCallback.call(root) if root.resolvedCallback?


  # ## Internal Helper


  buildAndRunFreakset: (opts) ->
    # Build the Freakset by copying the stack and passing the store
    opts.stack = opts.stack.slice() if opts.stack?
    freakset = Freakset.build(opts)
    freakset.next()
    return freakset


  pushToStack: (stack, object, opts) ->
    opts ||= {}

    if !opts.inFront && @isRunning()
      opts.inFront = yes
    else
      opts.inFront ||= no

    if opts.inFront is yes
      stack.unshift object
    else
      stack.push object


# Export the Build function
exports.Freakset = Freakset
